"""
CARPI GPS DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from redisdatabus.bus import TypedBusListener

KEY_BASE = 'carpi.gps.'


def build_key(type, name):
    global KEY_BASE
    return "{}{}{}".format(type, KEY_BASE, name)


# Raw GPS JSON String
KEY_RAW = build_key(TypedBusListener.TYPE_PREFIX_STRING, "RAW")
# Fix Mode (0=no mode, 1=no fix, 2=2D, 3=3D)
KEY_FIX_MODE = build_key(TypedBusListener.TYPE_PREFIX_INT, "fixmode")
# GPS Timestamp (UTC)
KEY_TIMESTAMP = build_key(TypedBusListener.TYPE_PREFIX_STRING, "timestamp")

# Latitude [°]
KEY_LATITUDE = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "latitude")
# Longitude [°]
KEY_LONGITUDE = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "longitude")
# Altitude [m]
KEY_ALTITUDE = build_key(TypedBusListener.TYPE_PREFIX_INT, "altitude")
# Course / Direction [°]
KEY_TRACK = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "track")
# Climb rate [m/s]
KEY_CLIMB = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "climb")

# Speed [m/s]
KEY_SPEED = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "speed")
# Speed [km/h]
KEY_SPEED_KMH = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "speed.kmh")
# Speed [mph]
KEY_SPEED_MPH = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "speed.mph")

# Longitude error [m]
KEY_EPX = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "epx")
# Latitude error [m]
KEY_EPY = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "epy")
# Altitude error [m]
KEY_EPV = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "epv")

# Course / Direction error [°]
KEY_EPD = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "epd")
# Speed error [m/s]
KEY_EPS = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "eps")
# Climb rate error [m/s]
KEY_EPC = build_key(TypedBusListener.TYPE_PREFIX_FLOAT, "epc")

# System Timestamp
KEY_SYS_TIMESTAMP = build_key(TypedBusListener.TYPE_PREFIX_STRING, "systimestamp")

# All Keys
ALL_KEYS = [
    KEY_RAW,
    KEY_FIX_MODE,
    KEY_TIMESTAMP,

    KEY_LATITUDE,
    KEY_LONGITUDE,
    KEY_ALTITUDE,
    KEY_TRACK,
    KEY_CLIMB,

    KEY_SPEED,
    KEY_SPEED_KMH,
    KEY_SPEED_MPH,

    KEY_EPX,
    KEY_EPY,
    KEY_EPV,

    KEY_EPD,
    KEY_EPS,
    KEY_EPC,

    KEY_SYS_TIMESTAMP,
]



"""
CARPI GPS DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from datetime import datetime
from logging import Logger
from time import sleep

from gpsdaemon.keys import *
from daemoncommons.daemon import Daemon
from carpicommons.log import logger
from carpicommons.errors import CarPiExitException
from redisdatabus.bus import BusWriter

from gps3 import agps3


class GpsdConnectionError(CarPiExitException):
    EXIT_CODE = 0xFA00

    def __init__(self):
        super().__init__(GpsdConnectionError.EXIT_CODE)


class GpsDaemon(Daemon):
    def __init__(self):
        super().__init__("GPS Daemon")
        self._log: Logger = None

    def _build_bus_writer(self):
        self._log.info("Connecting to Redis instance ...")
        return BusWriter(host=self._get_config('Redis', 'Host', '127.0.0.1'),
                         port=self._get_config_int('Redis', 'Port', 6379),
                         db=self._get_config_int('Redis', 'DB', 0),
                         password=self._get_config('Redis', 'Password', None))

    def startup(self):
        self._log = log = logger(self.name)
        log.info("Starting up %s ...", self.name)

        retries = 5

        bus = self._build_bus_writer()

        while True:
            sock = agps3.GPSDSocket()
            strm = agps3.DataStream()

            log.info("Connecting to GPSD...")
            sock.connect(self._get_config('GPSD', 'Host', agps3.HOST),
                         self._get_config_int('GPSD', 'Port', agps3.GPSD_PORT))
            sock.watch()

            log.info("Watching for new GPSD packages...")
            try:
                for d in sock:
                    if d:
                        strm.unpack(d)
                        di = self._build_messages(strm, d)
                        self._send_dict(bus, di)
            except OSError:
                if retries > 0:
                    log.error("Failed to fetch data from GPSD! Retrying %s more times", retries)
                    retries -= 1
                else:
                    raise GpsdConnectionError()

            sleep(5)

    def _build_messages(self, strm: agps3.DataStream, raw: str) -> dict:
        return {
            KEY_RAW: raw,
            KEY_FIX_MODE: self._get_int(strm.mode, 0),
            KEY_LATITUDE: self._get_float(strm.lat, None),
            KEY_LONGITUDE: self._get_float(strm.lon, None),
            KEY_TIMESTAMP: strm.time,
            KEY_SPEED: self._get_float(strm.speed),
            KEY_SPEED_KMH: self._get_float(strm.speed) * 3.6,
            KEY_SPEED_MPH: self._get_float(strm.speed) * 2.23694,
            KEY_TRACK: self._get_float(strm.track),
            KEY_CLIMB: self._get_float(strm.climb),
            KEY_EPX: self._get_float(strm.epx),
            KEY_EPY: self._get_float(strm.epy),
            KEY_EPV: self._get_float(strm.epv),
            KEY_EPD: self._get_float(strm.epd),
            KEY_EPS: self._get_float(strm.eps),
            KEY_EPC: self._get_float(strm.epc),
            KEY_SYS_TIMESTAMP: datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
        }

    def _get_int(self, v: any, default: int=0):
        return int(v) if v != 'n/a' else default

    def _get_float(self, v: any, default: float=0):
        return float(v) if v != 'n/a' else default

    def _send_dict(self, bus: BusWriter, d: dict):
        for key, value in d.items():
            bus.publish(key, str(value))

    def shutdown(self):
        self._log.info("Shutting down %s...", self.name)



"""
CARPI GPS DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from time import sleep
from typing import Any

from gpsdaemon.keys import ALL_KEYS, KEY_RAW
from redis import StrictRedis
from redisdatabus.bus import TypedBusListener


class GpsdClient(TypedBusListener):
    def __init__(self,
                 keys: list = ALL_KEYS,
                 name: str = None,
                 redis: StrictRedis = None,
                 host: str = '127.0.0.1',
                 port: int = 6379,
                 db: int = 0,
                 password: str = None):
        super().__init__(keys,
                         name,
                         redis,
                         host,
                         port,
                         db,
                         password)

    def _process_entry(self, msg: dict) -> (str, Any):
        return super()._process_entry(msg)


if __name__ == '__main__':
    a = ALL_KEYS
    a.remove(KEY_RAW)
    c = GpsdClient(keys=ALL_KEYS,
                   host='127.0.0.1',
                   port=6379,
                   db=0)
    c.start()
    try:
        while True:
            sleep(1)
    except (KeyboardInterrupt, SystemExit):
        print("Shutdown requested")
    finally:
        c.stop()



"""
CARPI GPS DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from daemoncommons.daemon import DaemonRunner

from gpsdaemon.daemon import GpsDaemon

if __name__ == '__main__':
    d = DaemonRunner('GPS_DAEMON_CFG', ['gps.ini', '/etc/carpi/gps.ini'])
    d.run(GpsDaemon())
